-- Browser for a tree of nodes.
--  * readonly; no editing
--  * nodes can be of 4 types: rows, cols, text or filler
--  * rows can contain cols/text/filler, cols can contain rows/text/filler
--  * text contains lines, lines can wrap and contain segments with inline styles
--  * filler occupies the full width under rows and the full height under cols. The other dimension is configurable.
--  * attributes: target, color, bg, border, font{size,bold,italic}, line_height
--  * computed attributes: x,y, width,height
--  * rows and cols can have attributes for themselves and also specify _default_ attributes for their descendants
--  * rows attributes: width, bg, border, default{color,bg,border,font,line_height}
--  * cols attributes: bg, border, default{color,bg,border,font,line_height}
--  * text attributes: width, color, bg, border, font, line_height
--  * line attributes: none (implicitly defaults to parent text)
--  * inline segment attributes: target, color, bg, font{bold,italic}. Inline segments can't change font size (it's too hard to make that always look nice)

-- A line without inline styles can be just a string:
--  'abc'
-- A line with inline styles is a Lua table containing a string and attributes for named segments:
--  {
--    'abc A@{def} B@{ghi}',
--    attrs={
--      A={bold=true},
--      B={color={1,0,0}, target='foo'},
--    }}
-- Segment names are alphanumeric and have to be preceded by whitespace or punctuation. While lines can contain Unicode, segment names must currently be ASCII.

local utf8 = require 'utf8'

local Font_height = 20

local Typographic_scale = {6,7,8,9,10,12,14,16,18,21,24,28,32,36,42,48,55,63,73,84,96}
local Font = {}

local Default_link_color = {0,0.7,1}

local Padding_x, Padding_y = 5, 5

local History = {}

function love.load(arg)
  love.graphics.setBackgroundColor(1,1,1)
  love.window.setMode(0,0)
  initialize_fonts()
  Document = load_document('screens/--help')
end

function initialize_fonts()
  for _,font_size in ipairs(Typographic_scale) do
    Font[font_size] = {
      -- bold?
      [false]={
        -- italic?
        [false]=love.graphics.newFont('fonts/Vera.ttf', font_size),
        [true]=love.graphics.newFont('fonts/Vera-Italic.ttf', font_size),
      },
      [true]={
        -- italic?
        [false]=love.graphics.newFont('fonts/Vera-Bold.ttf', font_size),
        [true]=love.graphics.newFont('fonts/Vera-Bold-Italic.ttf', font_size),
      },
    }
  end
end

function font_from_settings(font_settings)
  return Font[font_settings.size][bool(font_settings.bold)][bool(font_settings.italic)]
end

function bool(v)
  return not not v
end

function load_document(filename)
  table.insert(History, filename)
  local contents, err = love.filesystem.read(filename)
  if contents == nil then error('could not read '..filename..': '..err) end
  local chunk, err = load('return '..contents)
  if chunk == nil then error('could not parse '..filename ..': '..err) end
  document = chunk()
  assert(document, 'no value in '..filename)
  parse_all_inline_styles(document)
  compute_colors(document, --[[color]]{0,0,0}, --[[bg]]{1,1,1})  -- to compute colors of segments we need inline styles
  compute_fonts(document, {size=21})  -- to compute fonts of segments we need inline styles
  compute_widths(document)  -- to compute widths we need fonts for everything
  compute_line_heights(document, 35)
  compute_rects(document)  -- to know where to wrap lines we need widths
  compute_heights(document)  -- to compute heights we need to know where lines wrap as well as line heights
  return document
end

function parse_all_inline_styles(block)
  if block.type == 'rows' or block.type == 'cols' then
    for _,child in ipairs(block.data) do
      parse_all_inline_styles(child)
    end
  elseif block.type == 'text' then
    for i,line in ipairs(block.data) do
      if type(line) == 'string' then
        block.data[i] = {segments = {{data=line, attrs={}}}}
      else
        assert(line.segments == nil)
        block.data[i] = {segments = parse_inline_styles(line[1], line.attrs)}
      end
    end
  elseif block.type == 'filler' then
    -- nothing
  else
    assert(false, 'unknown type '..block.type)
  end
end

-- line is a table containing a string at position 1
-- populate inline styles into segments based on line.attrs
function parse_inline_styles(text, attrs)
  local segments = {}
  local offset = 1
  while offset <= #text do
    local styleStart, nameEnd, styleName = text:find("([%w_]+)@{", offset)
    if not styleStart then
      table.insert(segments, {data=text:sub(offset), attrs={}})
      break
    end
    if styleStart > offset then
      table.insert(segments, {data=text:sub(offset, styleStart-1), attrs={}})
    end
    local styleEnd = text:find("}", nameEnd)
    assert(styleEnd, "unbalanced inline style: "..text)
    local styledText = text:sub(nameEnd+1, styleEnd-1)
    table.insert(segments, {data=styledText, attrs=attrs[styleName] or {}})
    offset = styleEnd+1
  end
  return segments
end

-- postcondition: every segment of text has font set
-- default_font can be overridden by internal fonts
function compute_fonts(block, default_font)
  if block.type == 'rows' or block.type == 'cols' then
    for _,child in ipairs(block.data) do
      compute_fonts(child, (block.default and block.default.font) or default_font)
    end
  elseif block.type == 'text' then
    if default_font == nil then
      -- nothing
    elseif not block.font then
      block.font = deepcopy(default_font)
    elseif block.font and block.font.size then
      -- ignore default_font when changing font size
    elseif block.font and not block.font.size then
      -- inherit default font size
      block.font.size = default_font.size
    end
    assert(block.font, "could not determine font")
    assert(block.font.size, "could not determine font size")
    for _,line in ipairs(block.data) do
      for _,segment in ipairs(line.segments) do
        local font = deepcopy(block.font)
        if segment.attrs.font then
          assert(segment.attrs.font.size == nil, "inline styles can't change font size")
          for k,v in pairs(segment.attrs.font) do
            font[k] = v
          end
        end
        segment.attrs.font = font
      end
    end
  elseif block.type == 'filler' then
    -- nothing
  else
    assert(false, 'unknown type '..block.type)
  end
end

-- postcondition: every segment of text has color and bg set
-- default_color and default_bg can be overridden
function compute_colors(block, default_color, default_bg, default_border)
  if block.type == 'rows' or block.type == 'cols' then
    block.border = block.border or default_border
    for _,child in ipairs(block.data) do
      compute_colors(child,
          (block.default and block.default.color) or block.color or default_color,
          (block.default and block.default.bg) or block.bg or default_bg,
          (block.default and block.default.border) or block.border or default_border)
    end
  elseif block.type == 'text' then
    block.color = block.color or (block.target and Default_link_color) or default_color
    block.border = block.border or default_border
    block.bg = block.bg or default_bg
    for _,line in ipairs(block.data) do
      for _,segment in ipairs(line.segments) do
        segment.attrs.color = segment.attrs.color or (segment.attrs.target and Default_link_color) or block.color
        segment.attrs.bg = segment.attrs.bg or block.bg
      end
    end
  elseif block.type == 'filler' then
    block.border = block.border or default_border
    block.bg = block.bg or default_bg
  else
    assert(false, 'unknown type '..block.type)
  end
end

-- postcondition: every text descendant has line_height set
-- default_line_height can be overridden by internal line_heights
function compute_line_heights(block, default_line_height)
  if block.type == 'rows' or block.type == 'cols' then
    for _,child in ipairs(block.data) do
      compute_line_heights(child, (block.default and block.default.line_height) or default_line_height)
    end
  elseif block.type == 'text' then
    block.line_height = block.line_height or default_line_height
  elseif block.type == 'filler' then
    -- nothing
  else
    assert(false, 'unknown type '..block.type)
  end
end

-- postcondition: every rows/cols/text descendant has width set
-- if a node starts out with width set, it affects its descendants
-- if a node starts out without width set, it is affected by its descendants
-- imposed_width overrides any internal widths
-- default_font can be overridden by internal fonts
function compute_widths(block, imposed_width, default_font)
  if block.type == 'rows' then
    for _,row in ipairs(block.data) do
      compute_widths(row,
          imposed_width and (imposed_width-Padding_x*2) or block.width,
          (block.default and block.default.font) or default_font)
    end
    if imposed_width then
      block.width = imposed_width
    else
      block.width = 0
      for _,row in ipairs(block.data) do
        if row.type ~= 'filler' then
          block.width = math.max(block.width, row.width)
        end
      end
    end
    for _,row in ipairs(block.data) do
      if row.type == 'filler' then
        row.width = block.width
      end
    end
    block.width = block.width + Padding_x*2
  elseif block.type == 'cols' then
    assert(imposed_width == nil, "don't know how to split up width among cols")
    for _,col in ipairs(block.data) do
      compute_widths(col)
    end
    block.width = 0
    for _,col in ipairs(block.data) do
      block.width = block.width + col.width
    end
    block.width = block.width + Padding_x*2
  elseif block.type == 'text' then
    if imposed_width then
      block.width = imposed_width
    else
      -- compute width without wrapping
      block.width = 0
      for _,line in ipairs(block.data) do
        local width = 0
        for _,segment in ipairs(line.segments) do
          local font = font_from_settings(segment.attrs.font)
          width = width + font:getWidth(segment.data)
        end
        block.width = math.max(block.width, width)
      end
    end
    block.width = block.width + Padding_x*2
  elseif block.type == 'filler' then
    -- nothing
  else
    assert(false, 'unknown type '..block.type)
  end
end

function compute_rects(block)
  if block.type == 'rows' or block.type == 'cols' then
    for _,child in ipairs(block.data) do
      compute_rects(child)
    end
  elseif block.type == 'text' then
    compute_rects_for_text(block)
  elseif block.type == 'filler' then
    -- nothing
  else
    assert(false, 'unknown type '..block.type)
  end
end

function compute_rects_for_text(text)
  for _,line in ipairs(text.data) do
    compute_rects_for_line(line, text)
  end
end

function compute_rects_for_line(line, text)
  -- construct line without attrs as a single string for word-wrap calculations
  local line_text = {}
  for _, segment in ipairs(line.segments) do
    table.insert(line_text, segment.data)
  end
  line_text = table.concat(line_text)
  --
  local screen_lines = {}
  local curr_screen_line = {}
  local x, y = 0, 0
  local pos = 0
  for _,segment in ipairs(line.segments) do
    local font = font_from_settings(segment.attrs.font)
    for p,char in utf8chars(segment.data) do
      pos = pos+1
      assert(utf8_codepoint(line_text, pos) == char)
      local w = font:getWidth(char)
      if char:match('%s') then
        if should_word_wrap(font, text.width, x, line_text, pos) then
          table.insert(curr_screen_line,
            {x=x, y=y, width=w, height=text.line_height, data=char, attrs=segment.attrs})
          table.insert(screen_lines,
            {x=0, y=y, width=text.width, height=text.line_height, char_rects=curr_screen_line})
          curr_screen_line = {}
          x = 0
          y = y + text.line_height
        else
          table.insert(curr_screen_line, {x=x, y=y, width=w, height=text.line_height, data=char, attrs=segment.attrs})
          x = x + w
        end
      else
        if x+w > text.width then
          assert(pos > 1)
          table.insert(curr_screen_line,
            {x=x, y=y, width=text.width-x, height=text.line_height, attrs=segment.attrs})  -- filler
          table.insert(screen_lines,
            {x=0, y=y, width=text.width, height=text.line_height, char_rects=curr_screen_line})
          curr_screen_line = {}
          x = 0
          y = y + text.line_height
        else
          -- nothing
        end
        table.insert(curr_screen_line, {x=x, y=y, width=w, height=text.line_height, data=char, attrs=segment.attrs})
        x = x + w
      end
    end
  end
  table.insert(screen_lines,
    {x=0, y=y, width=text.width, height=text.line_height, char_rects=curr_screen_line})
  y = y + text.line_height
  line.rects = {x=0, y=0, width=text.width, height=y, screen_line_rects=screen_lines}
end

-- postcondition: every rows/cols/text descendant has height set
-- row/cols nodes never start out with height set
-- text height is sum of heights of screen lines
function compute_heights(block)
  if block.type == 'rows' then
    for _,row in ipairs(block.data) do
      compute_heights(row)
    end
    block.height = 0
    for _,row in ipairs(block.data) do
      block.height = block.height + row.height
    end
    block.height = block.height + Padding_y*2
  elseif block.type == 'cols' then
    for _,col in ipairs(block.data) do
      compute_heights(col)
    end
    block.height = 0
    for _,col in ipairs(block.data) do
      if col.type ~= 'filler' then
        block.height = math.max(block.height, col.height)
      end
    end
    for _,col in ipairs(block.data) do
      if col.type == 'filler' then
        col.height = block.height
      end
    end
    block.height = block.height + Padding_y*2
  elseif block.type == 'text' then
    block.height = 0
    for _,line in ipairs(block.data) do
      block.height = block.height + line.rects.height
    end
    block.height = block.height + Padding_y*2
  elseif block.type == 'filler' then
    -- nothing
  else
    assert(false, 'unknown type '..block.type)
  end
end

function love.draw()
  love.graphics.setColor(0,0,0)
  draw_block(Document, 10,10)
end

function draw_block(block, x,y)
  local orig_color, orig_font
  block.x, block.y = x, y
  if block.type == 'rows' then
    orig_color, orig_font = set_attrs(block, block, x,y, orig_color, orig_font)
    x, y = x + Padding_x, y + Padding_y
    for _,row in ipairs(block.data) do
      draw_block(row, x,y)
      y = y+row.height
    end
  elseif block.type == 'cols' then
    orig_color, orig_font = set_attrs(block, block, x,y, orig_color, orig_font)
    x, y = x + Padding_x, y + Padding_y
    for _,col in ipairs(block.data) do
      draw_block(col, x,y)
      x = x+col.width
    end
  elseif block.type == 'text' then
    orig_color, orig_font = set_attrs(block, block, x,y, orig_color, orig_font)
    x, y = x + Padding_x, y + Padding_y
    for _,line in ipairs(block.data) do
      for _,screen_line_rect in ipairs(line.rects.screen_line_rects) do
        x = block.x + Padding_x
        for _,char_rect in ipairs(screen_line_rect.char_rects) do
          set_attrs(char_rect, char_rect.attrs, x,y)
          if char_rect.data then
            love.graphics.print(char_rect.data, x,y)
          end
          char_rect.x, char_rect.y = x, y
          x = x + char_rect.width
        end
        y = y + screen_line_rect.height
      end
    end
  elseif block.type == 'filler' then
    orig_color = set_attrs(block, block, x,y, orig_color)
    if block.width then x = x + block.width end
    if block.height then y = y + block.height end
  else
    assert(false, 'unknown type '..block.type)
  end
  if orig_font then love.graphics.setFont(orig_font) end
  if orig_color then love.graphics.setColor(orig_color) end
end

function set_attrs(block, attrs, x,y, orig_color, orig_font)
  if attrs.bg then
    orig_color = orig_color or {love.graphics.getColor()}
    love.graphics.setColor(attrs.bg)
    love.graphics.rectangle('fill', x,y, block.width, block.height)
    love.graphics.setColor(orig_color)
  end
  if attrs.border then
    orig_color = orig_color or {love.graphics.getColor()}
    love.graphics.setColor(attrs.border)
    love.graphics.rectangle('line', x,y, block.width, block.height)
    love.graphics.setColor(orig_color)
  end
  if attrs.color then
    orig_color = orig_color or {love.graphics.getColor()}
    love.graphics.setColor(attrs.color)
  elseif attrs.target then
    orig_color = orig_color or {love.graphics.getColor()}
    love.graphics.setColor(Default_link_color)
  end
  if attrs.font then
    orig_font = orig_font or love.graphics.getFont()
    love.graphics.setFont(font_from_settings(attrs.font))
  end
  return orig_color, orig_font
end

function love.mousepressed(x,y, mouse_button)
  mouse_press_on_block(Document, x,y, mouse_button)
end

function mouse_press_on_block(block, x,y, mouse_button)
  if block.type == 'rows' then
    for _,row in ipairs(block.data) do
      if mouse_press_on_block(row, x,y, mouse_button) then
        return true
      end
    end
  elseif block.type == 'cols' then
    for _,col in ipairs(block.data) do
      if mouse_press_on_block(col, x,y, mouse_button) then
        return true
      end
    end
  elseif block.type == 'text' then
    if block.target and within_rect(block, x,y) then
      print('click on block '..block.target)
      Document = load_document('screens/'..block.target)
      return true
    end
    for _,line in ipairs(block.data) do
      for _,screen_line_rect in ipairs(line.rects.screen_line_rects) do
        for _,char_rect in ipairs(screen_line_rect.char_rects) do
          if within_rect(char_rect, x,y) then
            if char_rect.attrs.target then
              print('click on '..char_rect.attrs.target)
              Document = load_document('screens/'..char_rect.attrs.target)
              return true
            end
          end
        end
      end
    end
  elseif block.type == 'filler' then
    -- nothing
  else
    assert(false, 'unknown type '..block.type)
  end
end

function love.keypressed(key)
  if key == 'left' and alt_down() then
    if #History > 1 then
      table.remove(History)
      local filename = table.remove(History)
      Document = load_document(filename)
    end
  end
end

function alt_down()
  return love.keyboard.isDown('lalt') or love.keyboard.isDown('ralt')
end

function within_rect(rect, x,y)
  return within(x, rect.x, rect.x+rect.width)
    and within(y, rect.y, rect.y+rect.height)
end

function within(a, lo, hi)
  return a >= lo and a < hi
end

-- check if space character at pos/x is final possible candidate for word wrapping before right margin
function should_word_wrap(font, width, x, line, pos)
  if x < 0.8*width then return false end
  local offset = utf8_safe_offset(line, pos+1)
  local s = line:match('%S*', offset)
  local w = font:getWidth(s)
  return x+w > width
end

-- helper for debugging
function utf8_safe_offset(s, pos)
  if pos == 1 then return 1 end
  local result = utf8.offset(s, pos)
  if result == nil then
    assert(false, ('utf8_safe_offset(%d) called on a string of length %d (byte size %d); this is likely a failure to handle utf8\n\n^%s$\n'):format(pos, utf8.len(s), #s, s))
  end
  return result
end

-- create a new iterator for s which provides the index and UTF-8 bytes corresponding to each codepoint
function utf8chars(s)
  local next_pos = 1  -- in code points
  local next_offset = utf8.offset(s, next_pos)  -- in bytes
  return function()
    assert(next_offset)  -- never call the iterator after it returns nil
    local curr_pos = next_pos
    next_pos = next_pos+1
    local curr_offset = next_offset
    next_offset = utf8.offset(s, 2, next_offset)
    if next_offset == nil then return end
    local curr_char = s:sub(curr_offset, next_offset-1)
    return curr_pos, curr_char
  end
end

function utf8_codepoint(s, pos)
  local start_offset = utf8_safe_offset(s, pos)
  local end_offset = utf8_safe_offset(s, pos+1)
  assert(end_offset > start_offset, ('end_offset %d not > start_offset %d'):format(end_offset, start_offset))
  return s:sub(start_offset, end_offset-1)
end

-- https://stackoverflow.com/questions/640642/how-do-you-copy-a-lua-table-by-value/26367080#26367080
function deepcopy(obj, seen)
  if type(obj) ~= 'table' then return obj end
  seen = seen or {}
  if seen[obj] then return seen[obj] end
  local result = setmetatable({}, getmetatable(obj))
  seen[obj] = result
  for k,v in pairs(obj) do
    result[deepcopy(k, seen)] = deepcopy(v, seen)
  end
  return result
end
